/* Coded for sigMatch
 * Author: Ramakrishnan Kandhan (ramak@cs.wisc.edu)
 *
 * ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/*
 *  Compilation: gcc -Wall scan.c -o scan -lclamav
 *
 *  Author: Ramakrishnan Kandhan <ramak@cs.wisc.edu>
 *
 *  Scans a file using clamav and sigmatch+clamav and prints the
 *  time taken by both. Uses the sigTree data structure found in 
 *  sigTree.h. The bloom filter implemnetation can be found in 
 *  bloom_filter.h
 * 
 *  Note that this file may not work if the clamav implementation has been changed.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <clamav.h>

#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdint.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "sigTree.h"

#define MICRO_SECONDS 1000000
#define BUFFER_SIZE 1000

/*
 * Exit codes:
 *  0: clean
 *  1: infected
 *  2: error
 */

using namespace std;
using namespace sigMatch;

namespace {

FILE* fp;
int fp1;
struct sigRoot* root;
char* buff;
long int tbuff;
int len;
long int buffsize;

/* 
 * Calculate Time difference between two timeval variables  
 */

long timeDiff(struct timeval newerTime, struct timeval olderTime)
{
  long long diffs;
  diffs = (newerTime.tv_sec-olderTime.tv_sec)*MICRO_SECONDS+ 
  newerTime.tv_usec-olderTime.tv_usec;
  diffs /= 1000;
  return diffs;
}

}  // namespace

/*
 * Main
 */

int main(int argc, char **argv)
{
  int fd, ret;
  unsigned long int size = 0;
  unsigned int sigs = 0;
  long double mb;
  const char *virname;
  struct cl_engine *engine;

  struct timeval clamTimeBefore, clamTimeAfter;
  double clammbps;

  struct timeval sigTimeBefore, sigTimeAfter;
  double sigmbps;

  /* start loading the command line arguments */


  if(argc != 7) {
    printf("Usage: %s <file> <beta> <wordFile> <arraysdir> <bloom_size> <buffer_size>\n", argv[0]);
    return 2;
  }

  if((fd = open(argv[1], O_RDONLY)) == -1) {
	  printf("Can't open file %s\n", argv[1]);
    return 2;
  }

  len = atoi(argv[2]); //beta
  BLOOM_SIZE = atoi(argv[5]); //bloom filter size

  /* Load cl engine */

  if((ret = cl_init(CL_INIT_DEFAULT)) != CL_SUCCESS) {
	  printf("Can't initialize libclamav: %s\n", cl_strerror(ret));
    return 2;
  }

  if(!(engine = cl_engine_new())) {
    printf("Can't create new engine\n");
    return 2;
  }

  /* load all available databases from default directory */
  if((ret = cl_load(cl_retdbdir(), engine, &sigs, CL_DB_STDOPT)) != CL_SUCCESS) {
    printf("cl_load: %s\n", cl_strerror(ret));
    close(fd);
    cl_engine_free(engine);
    return 2;
  }

  /* build engine */
  if((ret = cl_engine_compile(engine)) != CL_SUCCESS) {
    printf("Database initialization error: %s\n", cl_strerror(ret));;
    cl_engine_free(engine);
    close(fd);
    return 2;
  }

  /* Set Max scan size and max file size */
	
  int r;
  uint64_t val64 = cl_engine_get_num(engine, CL_ENGINE_MAX_SCANSIZE, &r);
  cl_engine_set_num(engine, CL_ENGINE_MAX_SCANSIZE, val64);
  val64 = val64 * 10;
  val64 = cl_engine_get_num(engine, CL_ENGINE_MAX_FILESIZE, &r);
  printf ("val64 is %d\n" , (int) val64);
  val64 = val64 * 10;
  cl_engine_set_num(engine, CL_ENGINE_MAX_FILESIZE, val64);
	
  /* Start ClamAV scan */
  gettimeofday(&clamTimeBefore, NULL);

  /* scan file descriptor */
  if((ret = cl_scandesc(fd, &virname, &size, engine, CL_SCAN_STDOPT)) == CL_VIRUS) {
    printf("Virus detected: %s\n", virname);
  } else {
    if(ret == CL_CLEAN) {
      printf("No virus detected.\n");
	  } else {
      printf("Error: %s\n", cl_strerror(ret));
      cl_engine_free(engine);
      close(fd);
      return 2;
    }
  }
  
  gettimeofday(&clamTimeAfter, NULL);
  /* end ClamAV scan */

  long clamTime = timeDiff(clamTimeAfter,clamTimeBefore);
  close(fd);

  /* start sigMatch preprocessing */
  root = 	createTree(argv[3], len, argv[4]);
  fp = fopen (argv[1], "r");
  long numbuff = 0, thrown = 0;
  buffsize = atoi(argv[6]);
  buff = (char*) malloc (sizeof (char) *buffsize);

  char* newbuff = (char*) malloc (sizeof(char)*buffsize*100000);
  unsigned long int ind1 = 0; // long long int ind2 =0;
  tbuff = buffsize/2;

	/* start sigMatch + clamAV */
  gettimeofday(&sigTimeBefore, NULL);
  while (fread(buff, buffsize,1, fp)){
    numbuff++;
    if((ret = sigScanbuff(root,(unsigned char*) buff, buffsize, len/2)) == 0) {
      memcpy (newbuff + ind1, buff, buffsize);
	  ind1 = ind1+buffsize;
    }
    else {
      thrown ++;
    }
  }

  /* Clamav part */
  lseek(fp1, SEEK_SET, 0);

  if((ret = cl_sigscanbuff((unsigned char *) newbuff, ind1, 0, engine)) == CL_VIRUS) {
    printf("SigMatch has detected: %s\n", virname); 
  }
  gettimeofday(&sigTimeAfter, NULL);
  close (fp1);

  /* end sigMatch + ClamAV */

  long sigTime = timeDiff(sigTimeAfter,sigTimeBefore);
  fclose(fp);

	/* Caculate and print the stats */

  mb = size * (CL_COUNT_PRECISION / 1024) / 1024.0;
  clammbps =(double) (mb/(double) clamTime) * 1000 ;
  sigmbps =(double) (mb/(double) sigTime) * 1000 ;

  printf("Data scanned: %2.2Lf MB\n", mb);
  printf("Total Number of buffers: %ld\n", numbuff);
  printf("Total Number of buffers thrown out by sigMatch: %ld\n", thrown);
  printf("filter percentage: %f\n", ((double)thrown/(double)numbuff));
  printf("Time taken by ClamAV in milliseconds: %ld\n", clamTime);
  printf("Time taken by sigMatch + ClamAV in milliseconds: %ld\n", sigTime);
  printf("Scan Rate for ClamAV: %f MBps\n",clammbps);
  printf("Scan Rate for sigMatch + ClamAV: %f MBps\n",sigmbps);
  printf("Speedup is %f\n", (double) sigmbps/clammbps);

  /* free memory */
  cl_engine_free(engine);
  return ret == CL_VIRUS ? 1 : 0;
}
